unit VLgLcdG19;
{
Wrapper for the Logitech G19 Gaming keyboard display API
(c) Victor van Uitert, 2009
v.van.uitert [at] gmail.com

This software is provided 'as is'. Use it anyway you like.

Thanks to:
- lglcd by delphifreak
- lcdg15 by smurfynet
- Logitech for providing these fun tools


Possible changes and extensions to this class
- supply property to set wether UpdateLcd should be synchronous (now always async)
- implement message queue support (send keys and other events as Windows Messages)
- add 'pages', let client switch between multiple bitmaps
}
interface
uses
  SysUtils,
  Windows,
  Graphics,
  //Clipbrd,
  lglcd;

type
  // the buttons next to the Lcd
  TLgLcdButton = (
    lgbNone,   // received when the last pressed button was released
    lgbLeft,   // cursor
    lgbRight,  // cursor
    lgbUp,     // cursor
    lgbDown,   // cursor
    lgbOK,
    lgbCancel, // the 'back arrow', to the bottom-right of the cursor keys
    lgbMenu,
    lgbQuit); // not a physical button, you receive this when LCD Manager sends terminate_applet event


  // if you want to respond to the Lcd keys, set one of these event handlers
  // object method version of keypress callback
  TOnButtonEvent = procedure(Button : TLgLcdButton) of Object;
  // plain procedure version of keypress callback
  TOnButtonCB = procedure(Button : TLgLcdButton);

  // object method version of configure dialog request
  TOnConfigureEvent = procedure of Object;
  // plain procedure version of configure dialog request
  TOnConfigureCB = procedure;

  // used with DisplayPriority property, to affect your screen time in the
  // LCD Manager's display queue, relative to other LCD applications
  TLcdDisplayPriority = (
    ldpIdleNoShow = LGLCD_PRIORITY_IDLE_NO_SHOW, // Lowest priority, disable displaying. Use this priority when you dont have anything to show.
    ldpBackground = LGLCD_PRIORITY_BACKGROUND,   // Priority used for low priority items.
    ldpNormal     = LGLCD_PRIORITY_NORMAL,       // Normal priority, to be used by most applications most of the time.
    ldpAlert      = LGLCD_PRIORITY_ALERT);       // Highest priority. To be used only for critical screens, such as your CPU temperature is too high


  // main lcd wrapper class
  // this class is responsible for:
  // - contacting the Logitech LCD management runtime
  // - maintaining contact affter plug/unplug/standby/resume events
  // - creating a memory bitmap, whose canvas is available to draw on
  // - sending the contents of this bitmap to the lcd
  // - providing properties to link your own event handlers for handling
  //   keypresses and the LVD Manager's Configure button
  TVLgLcdG19 = class(TObject)
  private
    FApplicationName  : string;
    FIsAutostartable  : Boolean;
    FSupportConfigure : Boolean;
    FConnection : Integer;
    FDevice     : Integer; // 0: not (yet) connected, -1: disconnected
    FBitmap     : TBitmap;
    FLcdBitmap  : TlgLcdBitmapQVGAx32;
    FApplicationIsEnabled   : Boolean;
    FDisplayPriority        : TLcdDisplayPriority;
    FSynchronousUpdate      : Boolean;
    FOpenDeviceAttempts : Integer;
    FOnButton       : TOnButtonEvent;
    FOnButtonCB     : TOnButtonCB;
    FOnConfigure    : TOnConfigureEvent;
    FOnConfigureCB  : TOnConfigureCB;
    FForegroundApp  : Boolean;
    procedure LcdConnect;
    procedure LcdCheckConnection;
    procedure LcdDeviceLost;
    procedure LgLcdApiCheckRes(Res : DWORD);
    procedure LgDevCheckRes(Res : DWORD);
    procedure LcdOpenDevice;
    function GetCanvas: TCanvas;
    function GetDeviceConnected: Boolean;
    function ReceiveNotification(notificationCode: DWORD;
      notifyParam1, notifyParam2, notifyParam3, notifyParam4: DWORD): DWORD;
    procedure DoButton(Button : TLgLcdButton);
    procedure DoConfigure;
    procedure SetForegroundApp(const Value: Boolean);
    function GetButtonState: TLgLcdButton;
    function GetHeight: Integer;
    function GetWidth: Integer;
  public

    // Create will open connections to all Logitech 320x240 color displays
    // parameters:
    // ApplicationName will be displayed in the Logitech LCD Manager
    // IsAutostartable True will let the LCD Manager start the executable on boot
    // SupportConfigure will enable the Configure button in LCD Manager; you must
    // assign OnConfigure to receive callbacks for this!
    constructor Create(
      const ApplicationName : string;
      IsAutostartable : Boolean = False;
      SupportConfigure : Boolean = False);

    destructor Destroy; override;

    // asssign one of these with your own event handler to receive soft button
    // state changes
    // Note: return ASAP and make sure you are thread-safe; the handler is called
    // on a separate thread!
    property OnButton : TOnButtonEvent read FOnButton write FOnButton;
    property OnButtonCB : TOnButtonCB read FOnButtonCB write FOnButtonCB;

    // assign tone of these to be notified when user presses "Configure" button
    // in the LCD Manager application
    // Note: return ASAP and make sure you are thread-safe; the handler is called
    // on a separate thread!
    property OnConfigure : TOnConfigureEvent read FOnConfigure write FOnConfigure;
    property OnConfigureCB : TOnConfigureCB read FOnConfigureCB write FOnConfigureCB;

    // read this to know if user has disabled the app;
    // when False there is no need to update the canvas and lcd
    property ApplicationIsEnabled : Boolean read FApplicationIsEnabled;

    // read this to know if there is an active lcd connection
    // when False there is no need to update the canvas and lcd
    property DeviceConnected : Boolean read GetDeviceConnected;

    // Display priority affects the UpdateLcd procedure
    // warning: using ldpAlert too frequently will be blocked  by Lcd Manager!
    property DisplayPriority : TLcdDisplayPriority read FDisplayPriority write FDisplayPriority;
    // SynchronousUpdate affects the UpdateLcd procedure
    // Normally it's False, meaning the call returns before the device is actually updated
    // Set it to True to wait. This will prevent your app from 'overloading' the
    // lcd with too frequent updates.
    property SynchronousUpdate : Boolean read FSynchronousUpdate write FSynchronousUpdate;

    // set to True to make this application the foreground app on the lcd
    property ForegroundApp : Boolean read FForegroundApp write SetForegroundApp;

    // use the canvas to draw on the internal bitmap, call UpdateLcd when done
    property Canvas : TCanvas read GetCanvas;
    // as a convenience, the Lcd screen dimensions are offered (might be var in future)
    property Width : Integer read GetWidth;
    property Height : Integer read GetHeight;
    // fill canvas with provided color or default black
    procedure ClearBitmap(Color : TColor = clBlack);
    // use to make screendumps
    procedure SaveBitmap(const FileName : String);
    //clipboard support was disbled, it makes the app too big
    //procedure CopyBitmapToClipboard;

    // send the internal bitmap contents to the keyboard lcd
    procedure UpdateLcd;

    // returns the currently pressed lcd button
    // can be used as an alternative for setting the OnButton event handler
    property ButtonState : TLgLcdButton read GetButtonState;
  end;


implementation

const
  COpenAttempts = 10;
  ERROR_DEVICE_NOT_CONNECTED  = 1167; // missing in Windows.pas

function LcdTranslateSoftButton(Buttons : DWORD) : TLgLcdButton;
begin
    if Buttons = 0 then
      Result := lgbNone // al buttons are 'up'
    else if (Buttons and LGLCDBUTTON_LEFT) <> 0 then
      Result := lgbLeft
    else if (Buttons and LGLCDBUTTON_RIGHT) <> 0 then
      Result := lgbRight
    else if (Buttons and LGLCDBUTTON_UP) <> 0 then
      Result := lgbUp
    else if (Buttons and LGLCDBUTTON_DOWN) <> 0 then
      Result := lgbDown
    else if (Buttons and LGLCDBUTTON_OK) <> 0 then
      Result := lgbOK
    else if (Buttons and LGLCDBUTTON_CANCEL) <> 0 then
      Result := lgbCancel
    else if (Buttons and LGLCDBUTTON_MENU) <> 0 then
      Result := LgbMenu
    else
      Result := lgbNone; // unknown number, should never happen
end;

//******************************************************************************
{ Callback helpers }


function VLcdG19NotificationsCB(connection: Integer; pContext: Pointer;
                                notificationCode: DWORD;
                                notifyParam1, notifyParam2, notifyParam3, notifyParam4: DWORD
                                   ): DWORD; stdcall;
begin
  if (pContext <> nil) then
      Result := TVLgLcdG19(pContext).ReceiveNotification(
        notificationCode, notifyParam1, notifyParam2, notifyParam3, notifyParam4)
  else
    Result := 1;
end;

function VLcdG19OnSoftButtonsCB(device:Integer; dwButtons:DWORD; pContext:Pointer) : DWORD; stdcall;
{
In theory, a user could press more than one button at the same time.
Methinks, it is unlikely anyone would do this (it takes strength to press even one :-)
So, we only pass the single pressed button value.
}
var
  Button : TLgLcdButton;
begin
  if (pContext <> nil) then begin
      //TVLgLcdG19(pContext).OnSoftButtons(dwButtons);
      Button := LcdTranslateSoftButton(dwButtons);
      TVLgLcdG19(pContext).DoButton(Button);
      Result := 0;
  end else
    Result := 1;
end;

function VLcdG19OnConfigureCB(connection:integer; const pContext:Pointer) : DWORD; stdcall;
begin
  if (pContext <> nil) then begin
    TVLgLcdG19(pContext).DoConfigure;
    Result := 0;
  end else
    Result := 1;
end;

//******************************************************************************
//******************************************************************************
//******************************************************************************

{ TVLgLcdG19 }

procedure TVLgLcdG19.ClearBitmap(Color: TColor);
begin
  Canvas.Brush.Color := Color;
  Canvas.Brush.Style := bsSolid;
  Canvas.FillRect(Canvas.ClipRect);
end;

(* disabled; using the Clipbrd unit adds 300K to the executable!
procedure TVLgLcdG19.CopyBitmapToClipboard;
var
  LFormat : Word;
  LHandle : THandle;
  LPalette : HPALETTE;
begin
  FBitmap.SaveToClipBoardFormat(LFormat, LHandle, LPalette);
  ClipBoard.SetAsHandle(LFormat, LHandle);
end;
*)

constructor TVLgLcdG19.Create(const ApplicationName: string;
  IsAutostartable, SupportConfigure: Boolean);
begin
  FDisplayPriority := ldpNormal;
  FOpenDeviceAttempts := COpenAttempts; // we do not open the device yet, but will try when needed

  LgLcdApiCheckRes(lgLcdInit);

  FApplicationName := ApplicationName;
  FIsAutostartable := IsAutostartable;
  FSupportConfigure := SupportConfigure;
  LcdConnect; // can raise exception

  FBitmap := TBitmap.Create;
  FBitmap.PixelFormat := pf32bit;
  FBitmap.Width  := Width;
  FBitmap.Height := Height;
  ClearBitmap;
end;

destructor TVLgLcdG19.Destroy;
begin
  if FDevice <> 0 then
    lgLcdClose(FDevice);
  lgLcdDisconnect(FConnection);
  lgLcdDeInit;
  FreeAndNil(FBitmap);
end;

procedure TVLgLcdG19.LgDevCheckRes(Res: DWORD);
begin
  if (Res <> 0) then begin
    if (Res = ERROR_DEVICE_NOT_CONNECTED) then begin
      LcdDeviceLost;
    end else if (Res = ERROR_INVALID_PARAMETER) then begin
      // this also happens after resume form standby when FDevice is no longer valid!
      LcdDeviceLost;
      // this is a programmer error, so extreme punishment is needed :-)
      //raise Exception.Create('TVLgLcdG19: Invalid parameter in call to device');
    end;
  end;
end;

procedure TVLgLcdG19.DoButton(Button: TLgLcdButton);
begin
  if Assigned(OnButton) then
    OnButton(Button);
  if Assigned(OnButtonCB) then
    OnButtonCB(Button);
end;

procedure TVLgLcdG19.DoConfigure;
begin
  if Assigned(OnConfigure) then
    OnConfigure;
  if Assigned(OnConfigureCB) then
    OnConfigureCB;
end;

function TVLgLcdG19.GetButtonState: TLgLcdButton;
var
  Buttons : DWORD;
begin
  Result := lgbNone;
  if DeviceConnected then begin
    if lgLcdReadSoftButtons(FDevice, Buttons) = 0 then begin
      Result := LcdTranslateSoftButton(Buttons);
    end;
  end;
end;

function TVLgLcdG19.GetCanvas: TCanvas;
begin
  GetCanvas := FBitmap.Canvas;
end;

function TVLgLcdG19.GetDeviceConnected: Boolean;
begin
  if (FDevice = 0) and (FOpenDeviceAttempts > 0) then
    LcdOpenDevice;
  GetDeviceConnected := (FDevice <> 0);
end;

function TVLgLcdG19.GetHeight: Integer;
begin
  Result := LGLCD_QVGA_BMP_HEIGHT;
end;

function TVLgLcdG19.GetWidth: Integer;
begin
  Result := LGLCD_QVGA_BMP_WIDTH;
end;

procedure TVLgLcdG19.LcdCheckConnection;
begin
  // sometimes the connection itself becomes invalid, so we will try to reconnect
  if FConnection = 0 then
    LcdConnect;
end;

procedure TVLgLcdG19.LcdConnect;
var
  WName : WideString;
  ConnectContext : TlgLcdConnectContextExW;
begin
  ZeroMemory(@ConnectContext, sizeof (ConnectContext));
  WName := FApplicationName;
  ConnectContext.appFriendlyName := PWideChar(WName);
  if FIsAutostartable then
    Integer(ConnectContext.isAutostartable) := 1; // BOOL assign wil set to -1, which is not recognnized!
  if FSupportConfigure then begin
    ConnectContext.onConfigure.configCallback := @VLcdG19OnConfigureCB;
    ConnectContext.onConfigure.configContext := Self;
  end;
  ConnectContext.onNotify.notifyCallback := @VLcdG19NotificationsCB;
  ConnectContext.onNotify.notifyContext := Self;
  ConnectContext.dwAppletCapabilitiesSupported := LGLCD_APPLET_CAP_QVGA;
  // Create will fail with exception if the connection can't be established
  LgLcdApiCheckRes(lgLcdConnectExW(ConnectContext));
  FConnection := ConnectContext.connection;
end;

procedure TVLgLcdG19.LcdDeviceLost;
begin
  if FDevice <> 0 then
    lgLcdClose(FDevice); // must do this, otherwise lgLcdOpen will fail later!
  FDevice := 0;
  FOpenDeviceAttempts := COpenAttempts;
end;

procedure TVLgLcdG19.LcdOpenDevice;
var
    OpenByTypeContext: TlgLcdOpenByTypeContext;
    Res : DWORD;
begin
  if (FDevice = 0) then begin
    LcdCheckConnection;
    ZeroMemory(@OpenByTypeContext, sizeof (OpenByTypeContext));
    OpenByTypeContext.connection := FConnection;
    OpenByTypeContext.deviceType := LGLCD_DEVICE_QVGA;
    OpenByTypeContext.onSoftbuttonsChanged.softbuttonsChangedCallback := @VLcdG19OnSoftButtonsCB;
    OpenByTypeContext.onSoftbuttonsChanged.softbuttonsChangedContext := Self;
    Res := lgLcdOpenByType(OpenByTypeContext);
    //LgLcdApiCheckRes(Res);
    if (Res = 0) then begin
      FDevice := OpenByTypeContext.device;
      if ForegroundApp then
        SetForegroundApp(True);
    end else begin
      // when Res = ERROR_INVALID_PARAMETER it can mean
      // - there is no such device, or the lcd manager does not see it (yet)
      // - a device of this type has already been opened on this connection
      // - the connection is invalid
      // open failed, we will stop retrying, until event received
      if FOpenDeviceAttempts > 0 then
          Dec(FOpenDeviceAttempts);
//      if Res = ERROR_INVALID_PARAMETER then begin
//        lgLcdDisconnect(FConnection);
//        FConnection := 0;
//        LcdCheckConnection;
//      end;
    end;
  end;
end;

procedure TVLgLcdG19.LgLcdApiCheckRes(Res: DWORD);
const
  cBaseE = 'Logitech Lcd API error: ';
begin
  if Res <> 0 then begin
    if Res = ERROR_SERVICE_NOT_ACTIVE then
      raise Exception.Create(cBaseE + 'lgLcdInit has not been called yet.')
    else if Res = ERROR_INVALID_PARAMETER then
      raise Exception.Create(cBaseE + 'Either ctx or ctx->appFriendlyName are NULL.')
    else if Res = ERROR_FILE_NOT_FOUND then
      raise Exception.Create(cBaseE + 'LCDMon is not running on the system.')
    else if Res = ERROR_ALREADY_EXISTS then
      raise Exception.Create(cBaseE + 'The same client is already connected.')
    else if Res = RPC_X_WRONG_PIPE_VERSION then
      raise Exception.Create(cBaseE + 'LCDMon does not understand the protocol.')
    else
      raise Exception.Create(cBaseE + IntToStr(Res));
  end;
end;

function TVLgLcdG19.ReceiveNotification(notificationCode, notifyParam1,
  notifyParam2, notifyParam3, notifyParam4: DWORD): DWORD;
begin
  case notificationCode of
    // Notifications sent by LCD Manager to applets connected to it.
    LGLCD_NOTIFICATION_DEVICE_ARRIVAL   : begin
      if (notifyParam1 = LGLCD_DEVICE_QVGA) then begin
        LcdDeviceLost;  // try to open
      end;
    end;
    LGLCD_NOTIFICATION_DEVICE_REMOVAL   : begin
      if (notifyParam1 = LGLCD_DEVICE_QVGA) then begin
        LcdDeviceLost;
      end;
    end;
    LGLCD_NOTIFICATION_CLOSE_CONNECTION : begin
      LcdDeviceLost;
      FConnection := 0;
    end;
    LGLCD_NOTIFICATION_APPLET_DISABLED  : FApplicationIsEnabled := False;
    LGLCD_NOTIFICATION_APPLET_ENABLED   : FApplicationIsEnabled := True;
    LGLCD_NOTIFICATION_TERMINATE_APPLET : begin
      LcdDeviceLost;
      DoButton(lgbQuit);
    end;
  end;
  Result := 0;
end;

procedure TVLgLcdG19.SaveBitmap(const FileName: String);
begin
  FBitmap.SaveToFile(FileName);
end;

procedure TVLgLcdG19.SetForegroundApp(const Value: Boolean);
var
  Flag : Integer;
begin
  FForegroundApp := Value;
  if DeviceConnected and ApplicationIsEnabled then begin
    if Value then begin
      Flag := LGLCD_LCD_FOREGROUND_APP_YES;
      // bitmap must have been sent at least once, otherwise call will have no effect
      UpdateLcd;
    end else
      Flag := LGLCD_LCD_FOREGROUND_APP_NO;
    LgDevCheckRes(lgLcdSetAsLCDForegroundApp(FDevice, Flag));
  end;
end;

procedure TVLgLcdG19.UpdateLcd;
var
  P : ^Byte;
  B : ^Byte;
  Y : Integer;
  Priority: DWORD;
begin
  if not DeviceConnected then
    Exit;

  FLcdBitmap.hdr.Format := LGLCD_BMP_FORMAT_QVGAx32;
  // Pixel formats are identical; no per pixel conversion needed
  // Copy must be done per scanline, because internal format of Bitmap uses
  // reversed Y-order
  B := @FLcdBitmap.pixels;
  P := FBitmap.ScanLine[0];
  for Y := 0 to (LGLCD_QVGA_BMP_HEIGHT - 1) do begin
    //P := FBitmap.ScanLine[Y]; // <- slow!
    Move(P^, B^, LGLCD_QVGA_BMP_WIDTH * LGLCD_QVGA_BMP_BPP);
    Inc(B, LGLCD_QVGA_BMP_WIDTH * LGLCD_QVGA_BMP_BPP);
    Dec(P, LGLCD_QVGA_BMP_WIDTH * LGLCD_QVGA_BMP_BPP); // faster than calling ScanLine
  end;

  if SynchronousUpdate then
    Priority := LGLCD_SYNC_UPDATE(Ord(FDisplayPriority))
  else
    Priority := LGLCD_ASYNC_UPDATE(Ord(FDisplayPriority));

  LgDevCheckRes(lgLcdUpdateBitmap(
    FDevice,
    @FLcdBitmap,
    Priority));
end;

end.
